package com.me.pacman.state;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Input;
import com.badlogic.gdx.InputAdapter;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.g2d.BitmapFont;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.Vector2;
import com.me.pacman.PacDude;
import com.me.pacman.entity.*;
import com.me.pacman.entity.ai.ReturnToBase;
import com.me.pacman.level.Level;
import com.me.pacman.level.LevelTile;

import java.util.Random;

public class PlayState extends LevelState {

    public static final float FULL_SPEED = 9.375f;

    public static final Vector2[] GHOST_SPAWN_POINTS = {
        new Vector2(14f, 19.5f),
        new Vector2(14f, 16.5f),
        new Vector2(12f, 16.5f),
        new Vector2(16f, 16.5f),
    };

    public static final Direction[] GHOST_SPAWN_DIRS = {
            Direction.LEFT,
            Direction.UP,
            Direction.DOWN,
            Direction.DOWN,
    };

    private Texture levelBackground, winBackground;
    private BitmapFont font;

    private TextureRegion lifeSprite;

    public Random random;

    private long sirenId;

    private int pelletCount;
    private int pelletEatenCount;

    public int pelletsEatenSinceDeath;
    public boolean pelletsEatenSinceDeathCounterEnabled;

    private int score;
    private int lives;
    private int round;
    private boolean paused = false;

    private float readyTimer;
    private float gameOverTimer;
    private float newGameTimer;
    private float deathTimer;
    private float pointsTimer;
    private float roundClearedTimer;

    public float frightTimer;
    public float secondsSinceLastDot;

    private float chaseTimer;
    private float scatterTimer;
    private int scatterCount;

    public boolean chase;

    private boolean pacmanCaught;
    private int ghostsCaught;
    private Ghost lastGhostCaptured;

    public Pacman pacman;
    public Ghost[] ghosts;

    public PlayState(PacDude game) {
        super(game);
    }

    @Override
    public void setup() {
        levelBackground = game.assets.getLevelBackground();
        winBackground = game.assets.getLevelWinBackground();
        font = game.assets.getFont();
        lifeSprite = game.assets.pacman[2][1];
        Gdx.input.setInputProcessor(new Controller());

        newGame();
    }

    @Override
    public void render() {
        level.render(0, 16);

        game.assets.getFont().setColor(Color.WHITE);
        game.assets.getFont().draw(game.batch, "" + score, 40, 279);
        for (int i = 0; i < lives; i++) {
            game.batch.draw(lifeSprite, i * 16, 0);
        }

        if (pointsTimer > 0) {
            game.batch.draw(game.assets.points[0][ghostsCaught-1], (lastGhostCaptured.pos.x * 8) - 8, (lastGhostCaptured.pos.y * 8) + 8);
        } else {
            pacman.render(game.batch, 0, 16);
        }

        if (roundClearedTimer > 0 && roundClearedTimer <= 2) {
            // draw flashing level background
            game.batch.draw((int) (roundClearedTimer * 4) % 2 == 0? levelBackground : winBackground, 0, 16);
            return;
        } else {
            game.batch.draw(levelBackground, 0, 16);
        }

        if (pacman.alive && newGameTimer <= 0) {
            for (Ghost ghost : ghosts) {
                if (pointsTimer > 0 && ghost == lastGhostCaptured) {
                    continue;
                }
                ghost.render(game.batch, 0, 16);
            }
        }

        if (readyTimer > 0) {
            game.assets.getFont().setColor(Color.YELLOW);
            game.assets.getFont().draw(game.batch, "ready!", 92, 127);
        }

        if (gameOverTimer > 0) {
            game.assets.getFont().setColor(Color.RED);
            game.assets.getFont().draw(game.batch, "game  over", 78, 127);
        }

        if (paused) {
            game.assets.getFont().setColor(Color.YELLOW);
            game.assets.getFont().draw(game.batch, "paused", 90, 151);
        }
    }

    @Override
    public void update(float dt) {
        if (paused) {
            return;
        }

        if (PacDude.DEBUG) {
            // Fixed time step for debugger
            dt = 1/60f;
        }

        pointsTimer -= dt;
        if (pointsTimer > 0) {
            return;
        }

        if (roundClearedTimer > 0) {
            roundClearedTimer -= dt;
            if (roundClearedTimer <= 0) {
                newRound();
            }
            return;
        }

        if (gameOverTimer > 0) {
            gameOverTimer -= dt;
            if (gameOverTimer <= 0) {
                newGame();
            }
            return;
        }

        if (readyTimer > 0) {
            readyTimer -= dt;
            if (readyTimer <= 0) {
                startGame();
            } else if (newGameTimer <= 0) {
                return;
            }
        }

        if (newGameTimer > 0) {
            newGameTimer -= dt;
            if (newGameTimer <= 0) {
                setupGame();
            }
            return;
        }

        if (frightTimer > 0) {
            frightTimer -= dt;
            if (frightTimer <= 0) {
                for (Ghost ghost : ghosts) {
                    if (ghost.currentBehaviour instanceof ReturnToBase) continue;
                    ghost.caught = false;
                    ghost.currentBehaviour = ghost.chaseBehaviour;
                }
            }
        } else {
            // pause chase/scatter timer when frightened
            if (scatterTimer > 0) {
                scatterTimer -= dt;
                if (scatterTimer <= 0) {
                    chase = true;
                    if (scatterCount < 4) {
                        chaseTimer = 20f;
                    }
                }
            } else if (chaseTimer > 0) {
                chaseTimer -= dt;
                if (chaseTimer <= 0) {
                    chase = false;
                    scatterTimer = 7f;
                    scatterCount++;
                }
            }
        }

        secondsSinceLastDot += dt;
        if (secondsSinceLastDot >= 4) {
            // It's been 4 seconds since pacman last ate a dot, he's tryin' to avoid ghosts coming out!
            // We'll get him...
            for (int i = 1; i < 4; i++) {
                Ghost ghost = ghosts[i];
                if (ghost.inHouse) {
                    ghost.leaveHouse();
                    secondsSinceLastDot = 0;
                    break;
                }
            }
        }

        if (!pacmanCaught) {
            for (Ghost ghost : ghosts) {
                ghost.update(dt);

                if (ghost.onSameTile(pacman)) {
                    if (frightTimer > 0 && !ghost.caught) {
                        ghostCaught(ghost);
                    } else if (frightTimer <= 0 || !(ghost.currentBehaviour instanceof ReturnToBase)) {
                        pacmanCaught();
                        return;
                    }
                }
            }
        } else if (deathTimer <= 0) {
            if (pacman.alive) {
                game.assets.deathSound.play(1f);
            }

            pacman.alive = false;

            if (pacman.deathFrame == 13) {
                if (lives > 0) {
                    setupGame();
                    readyTimer = 2f;
                } else {
                    gameOverTimer = 2f;
                }
            }
        } else {
            deathTimer -= dt;
        }

        pacman.update(dt);
    }

    public void initializeField() {
        frightTimer = 0f;
        deathTimer = 0f;
        pacmanCaught = false;
        ghostsCaught = 0;
        lastGhostCaptured = null;
        secondsSinceLastDot = 0;

        if (chase) {
            chaseTimer = 20f;
        } else {
            scatterTimer = 7f;
        }

        random = new Random(897198256012865L);
        pacman = new Pacman(this, false);
    }

    public void newGame() {
        pelletCount = 0;
        pelletEatenCount = 0;
        score = 0;
        lives = 3;
        round = 0;

        newGameTimer = 2.3f;

        newRound();
    }

    public void newRound() {
        round++;

        level = new Level(game,"level");
        pelletCount = level.getTileCount(LevelTile.PELLET);
        pelletCount += level.getTileCount(LevelTile.POWER_PELLET);
        pelletsEatenSinceDeath = 0;
        pelletsEatenSinceDeathCounterEnabled = false;

        initializeField();
        spawnGhosts();

        chase = false;
        scatterCount = 1;

        game.assets.siren.stop(sirenId);
        if (round == 1) {
            readyTimer = 4.2f;
            game.assets.beginning_alt.play(1.0f);
        } else {
            readyTimer = 4.3f;
            game.assets.beginning.play(1.0f);
        }
    }

    public void setupGame() {
        lives--;
        initializeField();
        resetGhosts();
    }

    private void spawnGhosts() {
        ghosts = new Ghost[4];
        ghosts[0] = new Blinky(this, new Vector2(GHOST_SPAWN_POINTS[0]), GHOST_SPAWN_DIRS[0]);
        ghosts[1] = new Pinky(this, new Vector2(GHOST_SPAWN_POINTS[1]), GHOST_SPAWN_DIRS[1]);
        ghosts[2] = new Inky(this, new Vector2(GHOST_SPAWN_POINTS[2]), GHOST_SPAWN_DIRS[2]);
        ghosts[3] = new Clyde(this, new Vector2(GHOST_SPAWN_POINTS[3]), GHOST_SPAWN_DIRS[3]);
    }

    public void startGame() {
        pacman.moving = true;
        sirenId = game.assets.siren.loop(1.0f);
    }

    public void resetGhosts() {
        for (int i = 0; i < 4; i++) {
            if (ghosts[i] == null) continue;
            ghosts[i].pos = new Vector2(GHOST_SPAWN_POINTS[i]);
            ghosts[i].currDirection = GHOST_SPAWN_DIRS[i];
            ghosts[i].currentBehaviour = chase? ghosts[i].chaseBehaviour : ghosts[i].scatterBehaviour;
            ghosts[i].currentPath = null;
            ghosts[i].inHouse = i > 0;
        }
    }

    private void pelletEaten(float x, float y) {
        level.setTile(x, y, LevelTile.EMPTY);
        if (pelletEatenCount % 2 == 0) {
            game.assets.chomp_1.play(1.0f);
        } else {
            game.assets.chomp_2.play(1.0f);
        }
        pelletEatenCount++;
        pelletCount--;

        if (pelletsEatenSinceDeathCounterEnabled) {
            // Increase global dot counter when enabled
            pelletsEatenSinceDeath++;
        } else {
            // else increment appropriate ghost's individual counter
            for (Ghost ghost : ghosts) {
                // Increment the dot counter of the first encountered ghost in the house
                // Ghosts are ordered Blinky, Pinky, Inky, Clyde. Blinky's inHouse is always false,
                // Increase the dotCounter of only one ghost: the first of the above list still in the house.
                if (!ghost.inHouse) continue;
                ghost.dotCounter++;
                break;
            }
        }

        secondsSinceLastDot = 0;

        if (pelletCount == 0) {
            roundClearedTimer = 3f;
        }
    }

    public void eatPellet(float x, float y) {
        pelletEaten(x, y);
        score += 10;
    }

    public void eatPowerPellet(float x, float y) {
        pelletEaten(x, y);
        for (Ghost ghost : ghosts) {
            if (ghost.currentBehaviour instanceof ReturnToBase) continue;
            ghost.caught = false;
            ghost.currentBehaviour = ghost.frightBehaviour;
            ghost.currDirection = ghost.currDirection.getOpposite();
        }
        frightTimer = 6f;
        ghostsCaught = 0;
        score += 50;
    }

    private void pacmanCaught() {
        game.assets.siren.stop(sirenId);
        deathTimer = 1f;
        pacmanCaught = true;
        pacman.moving = false;
        pelletsEatenSinceDeath = 0;
        pelletsEatenSinceDeathCounterEnabled = true;
    }

    private void ghostCaught(Ghost ghost) {
        ghost.caught = true;
        ghost.currentBehaviour = new ReturnToBase(this);

        lastGhostCaptured = ghost;
        ghostsCaught++;
        score += ghostsCaught * 200;
        pointsTimer = 1f;
    }

    @Override
    public void dispose() {
        Gdx.input.setInputProcessor(null);
    }

    private final class Controller extends InputAdapter {

        @Override
        public boolean keyDown(int keycode) {
            switch (keycode) {
                case Input.Keys.UP:
                case Input.Keys.W:
                    pacman.setNextDirection(Direction.UP);
                    break;
                case Input.Keys.DOWN:
                case Input.Keys.S:
                    pacman.setNextDirection(Direction.DOWN);
                    break;
                case Input.Keys.LEFT:
                case Input.Keys.A:
                    pacman.setNextDirection(Direction.LEFT);
                    break;
                case Input.Keys.RIGHT:
                case Input.Keys.D:
                    pacman.setNextDirection(Direction.RIGHT);
                    break;
                case Input.Keys.ESCAPE:
                case Input.Keys.P:
                    paused = !paused;
                    break;
                case Input.Keys.N:
                    newGame();
                    break;
            }
            return super.keyDown(keycode);
        }

    }

}
